#include	"stdio.h"
#include	"realdeal.h"
#include	<stdarg.h>

long int ftell(FILE *f)
{
 long n;
 GetFPos(((FileData*)f)->fref, &n);
 return n;
}


long int fseek(FILE *f, int off, int mode)
{
 	return SetFPos(((FileData*)f)->fref, 
 		((mode == SEEK_SET) ? fsFromStart : ((mode == SEEK_CUR) ? fsFromMark: fsFromLEOF))
 		, off);
}

int	fscanf(FILE	*f, char	*format, ...)
{
	va_list	va;
	int		ret;
	
	va_start(va, format);
	ret = vfscanf(f, format, va);
	va_end(va);
	
	return ret;
}

int fprintf(FILE	*f, char	*format, ...)
{
	va_list	va;
	int		ret;
	
	va_start(va, format);
	ret = vfprintf(f, format, va);
	va_end(va);
	
	return ret;
}

int	vfprintf(FILE	*f, char	*format, va_list	args)
{
	char	buffer[512];
	int		ret;
	
	ret = vsprintf(buffer, format, args);
	fwrite(buffer, 1, strlen(buffer), f);
	
	return ret;
}

int	vfscanf(FILE	*f, char	*format, va_list	args)
{
	char	buffer[512];
	int		ret, used;
	long	oldpos;
	FileData	*fd = (FileData*)f;
	
	GetFPos(fd->fref, &oldpos);
	
	if(!fgets(buffer, 512, f))return 0;
	ret = vsscanf_with_count(buffer, &used, format, args);
	
	SetFPos(fd->fref, fsFromStart, oldpos + used);
	
	return ret;
}

int	fgetc(FILE	*f)
{
	char	c;
	FileData	*fd = (FileData*)f;
	
	if(fd->isconsole)return console_fgetc();
	
	if(fscanf(f, "%c", &c)<0)return -1;
	return c;
}

int	ungetc(int	c, FILE	*f)
{
	FileData	*fd = (FileData*)f;
	
	if(c == -1)return -1;
	if(fd->isconsole)
	{
		console_ungetc(c);
		return 0;
	}
	
	fd->pushback = c;
	return 0;
}

int	fwrite(void	*ptr, int	a, int	b, FILE	*f)
{
	long		len;
	FileData	*fd = (FileData*)f;
	
	len = a*b;
	if(fd->isconsole)
	{
		console_out_string(fd, ptr, len);
		return len;
	}
	//otherwise it's a real file
	fd->pushback = -1;
	
	if(FSWrite(fd->fref, &len, ptr))return -1;
	else return len;
}

int	fread(void	*ptr, int	a, int	b, FILE	*f)
{
	long		len, err, pbl = 0;
	FileData	*fd = (FileData*)f;
	
	len = a*b;
	if(len<0)return -1;
	if(len==0)return 0;
	
	if(fd->isconsole)
	{
		wait_for_input();
		console_in_string(ptr, len);
		return len;
	}
	
	//otherwise it's a real file
	if(fd->pushback != -1)
	{
		pbl = 1;
		*((char*)ptr)++ = fd->pushback;
		fd->pushback = -1;
		len--;
		if(len==0)return 1;
	}
	
	err = FSRead(fd->fref, &len, ptr);
	if(err != noErr && err != eofErr)return -1;
	else return len + pbl;
}

FILE	*fopen(char	*path, char	*perm)
{
	FileData	*fd;
	FSSpec		fsp;
	Str255		str;
	
	fd = (FileData*)NewPtr(sizeof(FileData));
	fd->isconsole = 0;
	fd->console = 0;
	
	strcpy((char*)str+1, path);
	str[0] = strlen(path);
	FSMakeFSSpec(0, 0, str, &fsp);
	fd->pushback = -1;
	if(FSpOpenDF(&fsp, fsRdWrPerm, &fd->fref) != noErr)
	{
		SysBeep(0);
		DisposePtr((char*)fd);
		return 0;
	}
	else
	{
		SetFPos(fd->fref, fsFromStart, 0);
		return (FILE*)fd;
	}
}

int	fclose(FILE	*f)
{
	FileData	*fd = (FileData*)f;
	
	if(fd->isconsole)return -1;	//right, close the console...
	FSClose(fd->fref);
	return 0;
}

int	feof(FILE	*f)
{
	FileData	*fd = (FileData*)f;
	long	end, cur;
	
	GetEOF(fd->fref, &end);
	GetFPos(fd->fref, &cur);
	return (cur>=end)?1:0;
}

char	*fgets(char  *ptr, int   max, FILE	*f)
{
	FileData	*fd = (FileData*)f;
	int i, count, fake;
	long	oldplace, newplace;
	
	if(fd->isconsole)
	{
		printf("ERROR: unsupported fgets on console\n");
		ptr[0] = '\n';
		ptr[1] = 0;
		return 0;
	}
	GetFPos(fd->fref, &oldplace);
	count = fread(ptr, 1, max, f);
	GetFPos(fd->fref, &newplace);
	fake = count - (newplace - oldplace);	//how many pushbacks were read
	if(count<0)
	{
		printf("ERROR: fread returned %d\n", count);
		ptr[0] = '\n';
		ptr[1] = 0;
		return 0;
	}
	if(count==0)return 0;
	for(i=0;i<count && ptr[i]!=10 && ptr[i]!=11 && ptr[i]!=13 && (ptr[i]!=-1);i++);
	//i is now the index of the first \n, hopefully.
	if(i>=(max-1))	//we ran out of room in the buffer before we hit a \n or an EOF
	{
		ptr[max-1] = 0;
		SetFPos(fd->fref, fsFromStart, oldplace + max - 1 - fake);
		return ptr;
	}
	else
	{
		SetFPos(fd->fref, fsFromStart, oldplace + i + 1 - fake);
		ptr[i] = '\n';
		i++;
		ptr[i] = 0;
		return ptr;
	}
}