/************************************************
 * Steven Stanek
 * September 01, 2001
 * yziformgzwrap.c
 * 
 * Implementation 0.0.0
 ***********************************************/

#include "yziformgzwrap.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "zlib.h"


static inline void initStream(z_stream *aStream);
static void fixNewLine(char *data);
static void* gzipMalloc(void *nothingness, unsigned int num, unsigned int size);
static void gzipFree(void *nothingness, void *freeme);


yzImage readCompressedFileImage(FILE *in)
{
	int comp_len, decomp_len;
	void *compressed;
	char *uncompressed;
	yzImage toReturn;
	
	compressed = readRestOfFile(in, &comp_len);
	uncompressed = (char*)decompressMe(compressed, comp_len, &decomp_len);
	
	toReturn = getImageFromString(uncompressed);
	free(compressed);
	free(uncompressed);
	
	return toReturn;
}

//Length is needed because /ns will occur within the string
yzImage processCompressedString(char *str)
{
	int decomp_len, comp_len;
	char *uncompressed;
	yzImage toReturn;
	
	assert(!strncmp(str, COMP_IMAGE_STR, strlen(COMP_IMAGE_STR)));
	str += strlen(COMP_IMAGE_STR);
	sscanf(str, "%d", &comp_len);
	
	assert(str = strchr(str, '\n')); 
	str ++;
	
	uncompressed  = (char*)decompressMe(str, comp_len, &decomp_len);
	
	toReturn = getImageFromString(uncompressed);
	free(uncompressed);
	
	return toReturn;
}


void *compressMe(void *data, int length, int *returnlen)
{
	z_stream aStream;
	void *out, *temp;
	int i, def_ret;
	
	initStream(&aStream);
	deflateInit(&aStream, Z_DEFAULT_COMPRESSION);
	
	aStream.next_in = data;
	aStream.avail_in = length;
	assert(out = aStream.next_out = malloc(STEP_SIZE));
	aStream.avail_out = STEP_SIZE;
	
	i=1;
	
	while(1)
	{
		def_ret = deflate(&aStream, Z_FINISH);
		
		if(def_ret == Z_STREAM_END)
			break;
		else if(def_ret == Z_OK)
		{
			assert(temp = malloc(STEP_SIZE * (i + 1)));
			
			memcpy(temp, out, aStream.total_out);
			free(out);
			
			out = temp;
			aStream.next_out = (unsigned char *)out + aStream.total_out;
			aStream.avail_out += STEP_SIZE;
			i++;
		}
		else
		{
			printf("Something's Wrong with deflation!\n");
			exit(55);
		}
	}
	
	*returnlen = aStream.total_out;
	
	if(deflateEnd(&aStream) != Z_OK)
	{
		printf("Error in ending deflation!!!!");
		exit(66);
	}
		
	return out;
}


void *decompressMe(void *data, int length, int *returnlen)
{
	z_stream aStream;
	void *out, *temp;
	int i, def_ret;
	
	initStream(&aStream);
	inflateInit(&aStream);
	
	aStream.next_in = data;
	aStream.avail_in = length;
	assert(out = aStream.next_out = malloc(length));
	aStream.avail_out = length;
	
	i=1;
	
	while(1)
	{
		def_ret = inflate(&aStream, Z_FINISH);
		
		if(def_ret == Z_STREAM_END)
			break;
		else if(def_ret == Z_OK || def_ret == Z_BUF_ERROR)
		{
			assert(temp = malloc(STEP_SIZE + aStream.total_out));

			memcpy(temp, out, aStream.total_out);
			free(out);
			
			out = temp;
			aStream.next_out = (unsigned char *)out + aStream.total_out;
			aStream.avail_out += STEP_SIZE;
			i++;
		}
		else
		{
			printf("Something's Wrong with inflation!\n");
			exit(55);
		}
	}
	
	*returnlen = aStream.total_out;
	inflateEnd(&aStream);
	
	fixNewLine((char*)out);
	
	return out;
}


static inline void initStream(z_stream *aStream)
{
	aStream->zalloc = gzipMalloc;
	aStream->zfree = gzipFree;
}


static void fixNewLine(char *data)
{
	for(data=data; data[0] != '\0'; data ++)
	{
		if(data[0] == '\r')
			data[0] = '\n';
	}
}


//Utilities
void *readRestOfFile(FILE *infile, int *returnLen)
{
	void *filedata;
	int startPos, endPos;
	int errcheck;
		
	startPos = ftell(infile);
	errcheck = fseek(infile, 0, SEEK_END);
	
	if(errcheck != 0)
	{
		printf("Fseek not working right, error code: %d", errcheck);
		exit(5211);
	}
	
	endPos = ftell(infile);
	errcheck = fseek(infile, startPos, SEEK_SET);
	
	if(errcheck != 0)
	{
		printf("Fseek not working right, error code: %d", errcheck);
		exit(5211);
	}
	
	assert(filedata = malloc(endPos - startPos));
	
	fread(filedata, 1, endPos-startPos, infile);
	*returnLen = endPos - startPos;
	return filedata;
}


//Wrappers
static void* gzipMalloc(void *nothingness, unsigned int num, unsigned int size)
{
	nothingness = nothingness;
	return malloc(size * num);
}


static void gzipFree(void *nothingness, void *freeme)
{
	nothingness = nothingness;
	free(freeme);
}