/*************************
 * Steven Stanek
 * yziformutil.c
 * Originally Written: March 18, 2001
 * Update to Version 0.0.1e completed 10_14_01
 *************************/
 
#include "yziformutil.h"
#include <math.h>

#ifndef bool
	#define bool int	
	#define	true 1
	#define	false 0
#endif	

#define COLOR_STEP_SIZE	512
#define AVG_DEF_LEN	 	26
#define AVG_STR_LEN		7
#define MAX_STR_LEN		15
#define MIN_OCCURANCES 	3
#define MAX_NUM_COLORS 	8096
#define BEGIN_DEFS_STR	"<Definitions>\n"
#define BEGIN_IMG_STR	"<Begin YZiForm Image Version 0.0.1e>\n"
#define END_IMG_STR		"<End Image>"

#define brev_32(x)  (x = (x<<24 & 0xFF000000) + (x<<8 & 0x00FF0000) + (x>>8 & 0x0000FF00) + (x>>24 & 0x000000FF))
#define brev_16(x)  (x = (x<<8 & 0xFF00) + (x>>8 & 0x00FF))

typedef struct
{
	int width, height;
	color_t **image_data;
} ImageData_t;


typedef struct llcol_t
{
	int num_occur;
	struct llcol_t *next;
	color_t dColor;
} llcol_t;


//Helpers for For colorArrtoImageFile
static void writeFile(ImageData_t *image, FILE *out);
static char* assembleDefsString(color_t *list, int numOccur, char ***IDs);
static char* assembleBodyString(ImageData_t *image, color_t *list, char **IDs, int numClrs);

//For assembleBodyString
static int 	findColorInArr(color_t *clrArr, color_t *findMe, int arrlen);
static void sortBySize(color_t *list, char **IDs, int numClrs);

//String Management
static void hexStr(char str[16], color_t *hexMe);
static inline char intToHex(int val);
static void expandStringTo(char **expandMe, int size);

//New Test Using Table
static int  assembleOccuranceList(ImageData_t *image, color_t **list, int **numOccur);
static int assembleDifColors(ImageData_t *image, color_t **list, int **numOccur);
static llcol_t *newFindInTable(color_t *findme, llcol_t **colors, int len);
static void newAddToList(color_t *add, llcol_t **list);
static int hashColor(color_t * hashme);


//For Assemble Occurance List
static void sortByNumOccurances(color_t *colorsList, int *occurList, int length);
static bool isTrans(color_t *aColor);

//Mindless image related utilities
static char* getNextID(char* last_ID);
static int  compareColors(color_t *color_a, color_t *color_b);


void colorArrtoImageFile(int width, int height, color_t** arr, FILE *out)
{
	ImageData_t image;
	image.width = width;
	image.height = height;
	image.image_data =arr;
	
	writeFile(&image, out);
}


static char* strdup2(char * todup)
{
	char * toreturn;
	
	assert(toreturn = malloc((strlen(todup) +1) * sizeof(char)));
	strcpy(toreturn, todup);
	return toreturn;
}



void colorArrtoBMP(color_t** image, FILE *out, int width, int height)
{
	unsigned int biWidth=width, biHeight=height;
	unsigned short bfType = 19778, bfReserved =0;
	unsigned int bfSize, bfOffset=54;
	unsigned short biPlanes=0, biBitCount = 24;
	unsigned int biSize = 40, biZero=0;
	int i, j;
	unsigned char r,g,b;

	bfSize = ((width*3) % 4 ==0) ? bfOffset+(3*width*height): bfOffset+(3*height*(width + 4 - (width*3) % 4));
	
	bfSize = (bfSize %4 == 0) ? (bfSize/4) *4:(bfSize/4) *4 +4;
	
	brev_16(bfType);
	brev_16(bfReserved);
	brev_32(bfSize);
	brev_32(bfOffset);
	brev_16(biPlanes);
	brev_16(biBitCount);
	brev_32(biSize);
	brev_32(biZero);
	brev_32(biWidth);
	brev_32(biHeight);
	
	fwrite(&bfType, 2, 1, out);
	fwrite(&bfSize, 4, 1, out);
	fwrite(&biZero, 4,1,out);
	fwrite(&bfOffset,4,1,out);
	fwrite(&biSize,4,1,out);
	fwrite(&biWidth,4,1,out);
	fwrite(&biHeight,4,1,out);
	fwrite(&biPlanes,2,1,out);
	fwrite(&biBitCount,2,1,out);
	fwrite(&biZero, 4,1,out);
	fwrite(&biZero, 4,1,out);
	fwrite(&biZero, 4,1,out);
	fwrite(&biZero, 4,1,out);
	fwrite(&biZero, 4,1,out);
	fwrite(&biZero, 4,1,out);
	
	for(i=height-1; i >= 0; i--)
	{
		for(j=0; j < width; j++)
		{
			r = (image[i][j].r) / 256;
			g = (image[i][j].g) / 256;
			b = (image[i][j].b) / 256;
			
			fwrite(&b,1,1,out);
			fwrite(&g,1,1,out);
			fwrite(&r,1,1,out);
		}
		
		for(j=0; width*3 % 4 != 0 &&  j < (4-(width*3)%4); j++)
		{
			r = 0;
			
			fwrite(&r,1,1,out);
		}
	}
}


void imageToBMP(yzImage dImage, FILE *out)
{
	int i;
	color_t** arr;
	
	arr=toColors(dImage);
	colorArrtoBMP(arr, out, getImageWidth(dImage), getImageHeight(dImage));
	
	for(i=0; i < getImageHeight(dImage); i++)
		free(arr[i]);
	free(arr);
}


color_t** readBMPFile(FILE *in, int *width, int *height)
{
	unsigned int intin;
	unsigned short shortin;
	unsigned char r, g, b;
	color_t **arr;
	int i, j;
	
	fread(&shortin,2,1,in);
	brev_16(shortin);
	
	if(shortin != 19778)
	{
		printf("Error: not a bitmap\n");
		return 0;
	}
	
	fread(&intin, 4,1,in);
	fread(&intin, 4,1,in);
	fread(&intin, 4,1,in);
	fread(&intin, 4,1,in);
	fread(&intin, 4,1,in);
	
	//Read width
	brev_32(intin);
	*width = intin;
	
	fread(&intin, 4,1,in);
	
	//Read height
	brev_32(intin);
	*height = intin;
	
	fread(&shortin,2,1,in);
	fread(&shortin,2,1,in);
	
	brev_16(shortin);
	
	if(shortin != 24)
	{
		printf("Error: Reader only supports 24 bit color\n");
		return 0;
	}
	
	fread(&intin, 4,1,in);
	
	brev_32(intin);
	
	if(intin != 0)
	{
		printf("This reader does not support compression\n");
		return 0;
	}
	
	fread(&intin, 4,1,in);
	fread(&intin, 4,1,in);
	fread(&intin, 4,1,in);
	fread(&intin, 4,1,in);
	fread(&intin, 4,1,in);
	
	assert(arr = malloc(*height * sizeof(color_t*)));
	
	for(i=0; i < *height; i++)
		assert(arr[i] = malloc(*width * sizeof(color_t)));
	
	for(i--; i >= 0; i--)
	{
		for(j=0; j < *width; j++)
		{
			fread(&b, 1, 1, in);
			fread(&g, 1, 1, in);
			fread(&r, 1, 1, in);
			
			arr[i][j].b = b * 256;
			arr[i][j].g = g * 256;
			arr[i][j].r = r * 256;
			arr[i][j].a=0;
		}
		
		//Read trailling zeros
		for(j=0; *width*3 % 4 != 0 &&  j < (4-(*width*3)%4); j++)
			fread(&r,1,1,in);
	}
	
	return arr;
}


void BMPfileToYZiFormFILE(FILE *bmpin, FILE *yziformout)
{
	int width, height, i;
	color_t **arr;
	
	arr=readBMPFile(bmpin, &width, &height);
	
	colorArrtoImageFile(width, height, arr, yziformout);
	
	for(i=0; i < height; i++)
		free(arr[i]);
	free(arr);
}


char* BMPtoYZiformStr(FILE *in)
{
	ImageData_t iDat;
	char *defs, *body, *full, **IDs;
	color_t *list;
	int numColors, *occurs, full_len, i;
	
	iDat.image_data=readBMPFile(in, &(iDat.width), &(iDat.height));
	
	numColors = assembleOccuranceList(&iDat, &list, &occurs);
	
	defs = assembleDefsString(list, numColors, &IDs);
	body = assembleBodyString(&iDat, list, IDs, numColors);
	
	free(iDat.image_data);
	free(list);
	free(occurs);
	
	for(i=0; i < numColors; i++)
		free(IDs[i]);
	free(IDs);
	
	full_len = strlen(defs) + strlen(body) + strlen(BEGIN_IMG_STR) + strlen(END_IMG_STR) + 1; //+1 for \0
	assert(full = malloc(sizeof(char) * full_len));
	
	strcpy(full, BEGIN_IMG_STR);
	strcat(full, defs);
	strcat(full, body);
	strcat(full, END_IMG_STR);
	
	free(defs);
	free(body);
	
	return full;
}


yzImage BMPtoyzImage(FILE *in)
{
	char * full;
	yzImage dImage;
	
	full=BMPtoYZiformStr(in);
	dImage = getImageFromString(full);
	free(full);
	return dImage;
}


/**********************************************************************************
 * 						HELPERS FOR colorArrtoImageFile
 **********************************************************************************/
static void writeFile(ImageData_t *image, FILE *out)
{
	color_t *colors;
	int *numOccur, totalcolors, i;
	char **IDs, *defs, *body;
	
	fprintf(out, "%s", BEGIN_IMG_STR);
	
	totalcolors = assembleOccuranceList(image, &colors, &numOccur);
	
	defs = assembleDefsString(colors, totalcolors, &IDs);
	fprintf(out, "%s", defs);
	free(defs);
	
	body = assembleBodyString(image, colors, IDs, totalcolors);
	fprintf(out, "%s", body);
	free(body);
	
	fprintf(out, "%s", END_IMG_STR);
	
	free(numOccur);
	free(colors);
	
	for(i=0; i < totalcolors; i++)
		free(IDs[i]);
	
	free(IDs);
	
}


static char* assembleDefsString(color_t *list, int numOccur, char ***IDs)
{
	char *defstr, *tempstr, *ID=0;
	int allocedSize, strused = 0, i;
	
	allocedSize = numOccur * AVG_DEF_LEN + strlen(BEGIN_DEFS_STR);
	assert(defstr = malloc(allocedSize * sizeof(char)));
	assert(*IDs = malloc(allocedSize * sizeof(char*)));
	
	strused = sprintf(defstr, "%s", BEGIN_DEFS_STR);
	
	for(i=0; i < numOccur; i++)
	{
		((*IDs)[i]) = ID = getNextID(ID);
		
		if(strlen(ID) == 1)
			strused += sprintf(defstr + strused, "%s = %d, %d, %d\n", ID, list[i].r, list[i].g, list[i].b);
		else
			strused += sprintf(defstr + strused, "\"%s\" = %d, %d, %d\n", ID, list[i].r, list[i].g, list[i].b);
		
		if(strused > allocedSize - 2 * AVG_DEF_LEN)
		{
			allocedSize += (numOccur - i +1) *AVG_DEF_LEN;
			assert(tempstr = malloc(allocedSize * sizeof(char)));
			strcpy(tempstr, defstr);	
			free(defstr);
			defstr = tempstr;
		}
	}
	
	return defstr;
}


static char* assembleBodyString(ImageData_t *image, color_t *list, char **IDs, int numClrs)
{
	char *dStr, hexedStr[16];
	int strAlloced, strWritten =0 , i, j, loc;
	
	sortBySize(list, IDs, numClrs);
	
	strAlloced = image->width * (image->height + 1) * AVG_STR_LEN;
	assert(dStr = malloc(sizeof(char) * strAlloced));
	
	strWritten += sprintf(dStr, "<Image %d, %d>\n", image->width, image->height);
	
	for(i=0; i < image->height; i++)
	{
		//String Management
		if(strAlloced - strWritten < image->width * MAX_STR_LEN +2)
		{
			strAlloced += image->width * MAX_STR_LEN +2;
			expandStringTo(&dStr, strAlloced);
		}
		
		for(j=0; j < image->width; j++)
		{
			loc = findColorInArr(list, &(image->image_data[i][j]), numClrs);
			
			if(loc < 0 && isTrans(&(image->image_data[i][j])))
				strWritten += sprintf(dStr + strWritten, "+");
				
			else if(loc < 0)
			{
				hexStr(hexedStr, &(image->image_data[i][j]));
				strWritten += sprintf(dStr + strWritten, "%s", hexedStr);
			}
			
			else
			{
				if(strlen(IDs[loc]) == 1)
					strWritten += sprintf(dStr + strWritten, "%s", IDs[loc]);
				else
					strWritten += sprintf(dStr + strWritten, "\"%s\"", IDs[loc]);
			}
		}
		
		strWritten += sprintf(dStr + strWritten, "\n");
	}
	sprintf(dStr + strWritten, "\0");
	
	return dStr;
}


//Returns Location in array, -1 if none exists
//Assumes Array has been sorted greatest to least (uses binary search)
static int 	findColorInArr(color_t *clrArr, color_t *findMe, int arrlen)
{
	color_t *top, *bottom, *middle;
	
	top = clrArr;
	bottom = clrArr + arrlen - 1;
	
	while(1)
	{
		if(top > bottom)
			return -1;
		
		middle = (bottom - top) /2 + top;
		
	//	if(middle == top || middle == bottom)
	//		return -1;
				
		switch(compareColors(middle, findMe))
		{
			case 0:		return middle - clrArr;
			case -1: 	bottom = middle -1;		break;
			case 1: 	top = middle + 1;				break;
			default: 	assert(1 == 0);
		}
	}
}


//Try qsort later
static void sortBySize(color_t *list, char **IDs, int numClrs)
{
	int i, j, largestLoc;
	color_t tempClr;
	char *tempStr;
	
	for(i=0; i < numClrs; i++)
	{
		largestLoc = i;
		
		for(j=i +1; j < numClrs; j++)
			if(compareColors(&(list[j]), &(list[largestLoc])) > 0)
				largestLoc = j;
	
		tempStr = IDs[largestLoc];
		tempClr = list[largestLoc];
		IDs[largestLoc] = IDs[i];
		list[largestLoc] = list[i];
		IDs[i] = tempStr;
		list[i] = tempClr;
	}
}


//Takes array of length 16 or greater 
//LENGTH MUST BE 16 or GREATER!!!!
static void hexStr(char str[16], color_t *hexMe)
{
	str[0] = '\"';
	str[1] = '#';
	str[2] = intToHex(hexMe->r / 4096);
	str[3] = intToHex((hexMe->r / 256) % 16);
	str[4] = intToHex((hexMe->r / 16) % 16);
	str[5] = intToHex(hexMe->r % 16);
	str[6] = intToHex(hexMe->g / 4096);
	str[7] = intToHex((hexMe->g / 256) % 16);
	str[8] = intToHex((hexMe->g / 16) % 16);
	str[9] = intToHex(hexMe->g % 16);
	str[10] = intToHex(hexMe->b / 4096);
	str[11] = intToHex((hexMe->b / 256) % 16);
	str[12] = intToHex((hexMe->b / 16) % 16);
	str[13] = intToHex(hexMe->b % 16);
	str[14] = '\"';
	str[15] = '\0';
}


static inline char intToHex(int val)
{
	//assert(val > 0 && val < 16);
	
	if (val < 10)
		return '0' + val;
	else
		return 'A' + val - 10;
	
		
}

//Increases size of string to size 
//Note if size < strlen, it will fail assertion
static void expandStringTo(char **expandMe, int size)
{
	char *newStr;
	
	assert(size > strlen(*expandMe));
	assert(newStr = malloc(sizeof(char) * size));
	strcpy(newStr, *expandMe);
	free(*expandMe);
	*expandMe = newStr;
}


static int assembleOccuranceList(ImageData_t *image, color_t **list, int **numOccur)
{
	int numCol;
	
	numCol = assembleDifColors(image, list, numOccur);
	
	sortByNumOccurances(*list, *numOccur, numCol);
	
	return numCol;
}


static int assembleDifColors(ImageData_t *image, color_t **list, int **numOccur)
{
	llcol_t **colors, *start, *last, *pres;
	int len, i, j, totalcolors = 0, col_count = 0;
	
	len = sqrt(image->width * image->height);
	
	assert(colors = malloc(sizeof(llcol_t *) * len));
	
	for(i=0; i < len; i++)
		colors[i] = 0;
	
	
	//Assemble Table
	for(i=0; i < image->height; i++)
	{
		for(j=0; j < image->width; j++)
		{
			pres = newFindInTable(&(image->image_data[i][j]), colors, len);
			
			if(!pres)
				newAddToList(&(image->image_data[i][j]), &(colors[hashColor(&(image->image_data[i][j]))%len]));
			else
				pres->num_occur ++;
		}
	}
	
	//Weed out nominal colors & count total colors
	for(i=0; i < len; i++)
	{
		last = 0;
		for(start = colors[i]; start; /*start = start->next*/)
		{
			if(start->num_occur < MIN_OCCURANCES)
			{
				if(last == 0)
					colors[i] = start->next;
				else
					last->next = start->next;
				pres = start;
				start= pres->next;
				free(pres);
			}
			else
			{
				totalcolors ++;
				last = start;
				start = start->next;
			}
		}	
	}
	
	assert(*list = malloc(sizeof(color_t) * totalcolors));
	assert(*numOccur = malloc(sizeof(int) * totalcolors));
	
	//Turn into list
	for(i=0; i < len; i++)
	{
		for(start = colors[i]; start;)
		{
			(*list)[col_count] = (start->dColor);
			(*numOccur)[col_count] = (start->num_occur);
			col_count ++;
			assert(col_count <= totalcolors);
			pres = start;
			start = start->next;
			free(pres);
		}
	}
	
	assert(col_count == totalcolors);
	
	return totalcolors;
}


static llcol_t *newFindInTable(color_t *findme, llcol_t **colors, int len)
{
	llcol_t *start;
	
	start = colors[hashColor(findme) % len];
	
	for(start = colors[hashColor(findme) % len]; start; start = start->next)
	{
		if(!compareColors(findme, &(start->dColor)))
			return start;
	}
 	
 	return 0;
}


static void newAddToList(color_t *add, llcol_t **list)
{
	llcol_t *newllct;
	
	assert(newllct = malloc(sizeof(llcol_t)));
	newllct->next = *list;
	newllct->dColor = *add;
	newllct->num_occur = 1;
	*list = newllct;
}


static int hashColor(color_t * hashme)
{
	return (hashme->r / 256 + (hashme->g /256) * 256 + (hashme->b / 256) * 256 * 256);
}

//Right now selection sort, may become qsort later
static void sortByNumOccurances(color_t *colorsList, int *occurList, int length)
{
	int i, j, largestAddress, tempInt;
	color_t tempCol;
	
	for(i=0; i < length; i++)
	{
		largestAddress = i;
		
		//Find Largest
		for(j=i; j < length; j++)
			if(occurList[j] > occurList[largestAddress])
				largestAddress = j;
		
		tempInt = occurList[i];
		tempCol = colorsList[i];
		occurList[i] = occurList[largestAddress];
		colorsList[i] = colorsList[largestAddress];
		occurList[largestAddress] = tempInt;
		colorsList[largestAddress] = tempCol;
	}
}


static bool isTrans(color_t *aColor)
{
	if(aColor->a == TRANS_A)
		return true;
	else
		return false;
}

//Pass NULL for beginning
static char* getNextID(char* last_ID)
{	
	int len, i;
	char *newStr;
	
	//If NULL return the starting string, the # 0
	if(!last_ID)
		return (char *) strdup2("0");
		
	len = strlen(last_ID);
		
	//Add another digit
	if(last_ID[len-1] == 'z' && last_ID[0] == 'z')
	{
		assert(newStr = malloc((len+2) *sizeof(char)));
			
		for(i=0; i < len +1; i++)
			newStr[i] = '0';
			
		newStr[len+1] = '\0';
		
		return newStr;
	}
	
	//If last digit = 'z' - roll back next non 'z' digit and replace previous
	// 'z's with '0's
	else
	{
		newStr = strdup2(last_ID);
		
		for(i=len-1; i >= 0; i--)
		{
			if(newStr[i] == 'z')
				newStr[i] = '0';
			
			else
			{	if(newStr[i] != ';')
					newStr[i] ++;
				else
					newStr[i] +=2;
				break;
			}
		}
				
		return newStr;
	}
}


static int compareColors(color_t *color_a, color_t *color_b)
{
	if(color_a->a == color_b->a == TRANS_A)
		return 0;
	
	if(color_a->r > color_b->r)
		return 1;
	else if(color_a->r < color_b->r)
		return -1;
	
	if(color_a->g > color_b->g)
		return 1;
	else if(color_a->g < color_b->g)
		return -1;
	
	if(color_a->b > color_b->b)
		return 1;
	else if(color_a->b < color_b->b)
		return -1;
		
	return 0;
}